#    Formatter.py
#
#    ------------------------------------------------------------
#    Copyright 2002, 2004 by Samuel Reynolds. All rights reserved.
#
#    Permission to use, copy, modify, and distribute this software and its
#    documentation for any purpose and without fee is hereby granted,
#    provided that the above copyright notice appear in all copies and that
#    both that copyright notice and this permission notice appear in
#    supporting documentation, and that the name of Samuel Reynolds
#    not be used in advertising or publicity pertaining to distribution
#    of the software without specific, written prior permission.
#
#    SAMUEL REYNOLDS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
#    INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
#    EVENT SHALL SAMUEL REYNOLDS BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
#    CONSEQUENTIAL DAMAGES, OR FOR ANY DAMAGES WHATSOEVER RESULTING FROM
#    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
#    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
#    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#    ------------------------------------------------------------
"""
Formatters for converting and validating data values.
"""

import os, sys, errno, copy, re
import six
import wx


class Formatter(object):
    """
    Formatter/validator for data values.
    """
    def __init__(self, *args, **kwargs):
        pass

    # Default (dummy) validate routine
    def validate(self, str_value):
        """
        Return true if value is valid for the field.
        value is a string from the UI.
        """
        return True

    # Default (dummy) format routine
    def format(self, value):
        """Format a value for presentation in the UI."""
        if value is None:
            return ''
        return str(value)

    # Default (dummy) coerce routine
    def coerce(self, str_value):
        """Convert a string from the UI into a storable value."""
        return str_value


class ChoiceFormatter(Formatter):
    """
    Formatter for choice data values.
    """
    def __init__(self, mapping, sort=None, *args, **kwargs):

        super(ChoiceFormatter, self).__init__(*args, **kwargs)
        # if mapping is a list, convert it to a dict
        if isinstance(mapping, list):
            mapping = {v: str(v) for v in mapping}
        self.mapping = mapping
        self.mapping_reverse = {v: k for k, v in six.iteritems(self.mapping)}
        self.sort = sort

    def validValues(self):
        """
        Return list of valid value (id,label) pairs.
        """
        items = list(six.iteritems(self.mapping))
        if self.sort:

            def sort_by_label(item):
                return item[1]

            key = None
            if self.sort == 'label':
                key = sort_by_label
            items.sort(key=key)
        return items

    def validate(self, str_value):
        """
        Return true if value is valid for the field.
        value is a string from the UI.
        """
        vv = [s for i, s in self.validValues()]
        return str_value in vv

    def format(self, value):
        """Format a value for presentation in the UI."""
        return self.mapping[value]

    def coerce(self, str_value):
        """Convert a string from the UI into a storable value."""
        return self.mapping_reverse[str_value]


class EnumFormatter(Formatter):
    """
    Formatter for enumerated (EnumType) data values.
    """
    def __init__(self, enumeration, sort=None, *args, **kwargs):

        super(EnumFormatter, self).__init__(*args, **kwargs)

        self.enumeration = enumeration
        self.sort = sort

    def validValues(self):
        """
        Return list of valid value (id,label) pairs.
        """
        items = copy.copy(self.enumeration.items())
        if self.sort:

            def sort_by_label(item):
                return item[1]

            key = None
            if self.sort == 'label':
                key = sort_by_label
            items.sort(key=key)
        return items

    def validate(self, str_value):
        """
        Return true if value is valid for the field.
        value is a string from the UI.
        """
        vv = [s for i, s in self.validValues()]
        return str_value in vv

    def format(self, value):
        """Format a value for presentation in the UI."""
        return self.enumeration[value]

    def coerce(self, str_value):
        """Convert a string from the UI into a storable value."""
        return getattr(self.enumeration, str_value)


class FormatterMeta(type):
    """
    Metaclass for subclasses of Formatter.

    Each instance class MUST define either validate(self, value) method
    or re_validation regular expression string.

    If the latter, the validate method will be autogenerated.

    If re_validation is defined, validate method is overridden to
    validate the string value against the regular expression.

    If re_validation_flags is defined, the flags will be
    used when the re_validation regular expression string is compiled.

    Each instance class MAY define:
    -    format(self, value) method.
    -    coerce(self, value) method.
    """
    def __new__(cls, classname, bases, classdict):
        newdict = copy.copy(classdict)

        # Generate __init__ method
        # Direct descendants of Formatter automatically get __init__.
        # Indirect descendants don't automatically get one.
        if '__init__' not in newdict:
            if Formatter in bases:

                def __init__(self, *args, **kwargs):
                    Formatter.__init__(self, *args, **kwargs)
                    initialize = getattr(self, 'initialize', None)
                    if initialize:
                        initialize(*args, **kwargs)

                newdict['__init__'] = __init__
            else:

                def __init__(self, *args, **kwargs):
                    super(self.__class__, self).__init__(*args, **kwargs)
                    initialize = getattr(self, 'initialize', None)
                    if initialize:
                        initialize(*args, **kwargs)

                newdict['__init__'] = __init__

        # Generate validate-by-RE method if specified
        re_validation = newdict.get('re_validation', None)
        if re_validation:
            # Override validate method
            re_validation_flags = newdict.get('re_validation_flags', 0)
            newdict['_re_validation'] = re.compile(re_validation,
                                                   re_validation_flags)

            def validate(self, str_value):
                return self._re_validation.match(str_value) != None

            newdict['validate'] = validate

        # Delegate class creation to the expert
        return type.__new__(cls, classname, bases, newdict)


class ObjectIdFormatter(six.with_metaclass(FormatterMeta, Formatter)):
    """
    Object ID is assumed to be a large (32 bit?) unsigned integer.
    """
    re_validation = '^[0-9]+$'

    def coerce(self, str_value):
        if str_value:
            return int(str_value)
        return str_value


class StringFormatter(six.with_metaclass(FormatterMeta, Formatter)):
    pass


class AlphaFormatter(StringFormatter):
    """Alphabetic characters only."""
    re_validation = '^[a-zA-Z]*$'


class AlphaNumericFormatter(StringFormatter):
    """Alphanumeric characters only."""
    re_validation = '^[a-zA-Z0-9]*$'


class EmailFormatter(StringFormatter):
    """Internet email addresses (more or less)."""
    # This regex does not match all legal email addresses, but
    # it does a pretty good job.
    # Strangely enough, '/' is legal in email addresses.
    # However, I've never seen it used, so I prefer to leave it out.
    _re_subs = {
        'sub1': r'[a-zA-Z~_-][a-zA-Z0-9_:~-]*',
        'sub2': r'(\.[a-zA-Z0-9_:~-]+)*',
        'sfx': r'\.[a-zA-Z]{2,3}'
    }
    re_validation = '^%(sub1)s%(sub2)s[@]%(sub1)s%(sub2)s%(sfx)s$' % _re_subs


class MoneyFormatter(StringFormatter):
    """Assumes decimal money, but doesn't assume currency."""
    re_validation = '^(([0-9]+([.][0-9]{2})?)|([0-9]*[.][0-9]{2}))$'


class IntFormatter(six.with_metaclass(FormatterMeta, Formatter)):
    """Signed or unsigned integer."""

    #re_validation = '^[-+]?[0-9]+$'

    def __init__(self, min_val=0, max_val=2**31 - 1, **kwargs):
        super(IntFormatter, self).__init__(min_val, max_val, **kwargs)

        self.max_val = max_val
        self.min_val = min_val

    def validate(self, str_value):
        try:
            v = self.coerce(str_value)
            if self.max_val is not None and v > self.max_val:
                return False
            if self.min_val is not None and v < self.min_val:
                return False
            return True
        except:
            return False

    def coerce(self, str_value):
        if str_value:
            return int(str_value)
        return str_value


class HexFormatter(IntFormatter):
    def format(self, value):
        return hex(value)

    def coerce(self, str_value):
        return int(str_value, 16)


class BinFormatter(IntFormatter):
    def format(self, value):
        return bin(value)

    def coerce(self, str_value):
        return int(str_value, 2)


class BoolFormatter(IntFormatter):
    def format(self, value):
        return str(bool(value))

    def coerce(self, str_value):
        str_value = str_value.strip()
        if str_value.lower() == 'true':
            return 1
        return 0


class Int8Formatter(IntFormatter):
    def __init__(self, min_val=-2**7, max_val=2**7 - 1, **kwargs):
        super(Int8Formatter, self).__init__(min_val, max_val, **kwargs)


class Int16Formatter(IntFormatter):
    def __init__(self, min_val=-2**15, max_val=2**15 - 1, **kwargs):
        super(Int16Formatter, self).__init__(min_val, max_val, **kwargs)


class Int24Formatter(IntFormatter):
    def __init__(self, min_val=-2**23, max_val=2**23 - 1, **kwargs):
        super(Int24Formatter, self).__init__(min_val, max_val, **kwargs)


class Int32Formatter(IntFormatter):
    def __init__(self, min_val=-2**31, max_val=2**31 - 1, **kwargs):
        super(Int32Formatter, self).__init__(min_val, max_val, **kwargs)


class Int64Formatter(IntFormatter):
    def __init__(self, min_val=-2**63, max_val=2**63 - 1, **kwargs):
        super(Int64Formatter, self).__init__(min_val, max_val, **kwargs)


class UIntFormatter(IntFormatter):
    """Unsigned integer."""
    def __init__(self, min_val=None, max_val=None, **kwargs):
        super(UIntFormatter, self).__init__(min_val, max_val, **kwargs)

        self.min_val = max(0, min_val)


class UInt8Formatter(UIntFormatter):
    def __init__(self, min_val=0, max_val=2**8 - 1, **kwargs):
        super(UInt8Formatter, self).__init__(min_val, max_val, **kwargs)


class UInt16Formatter(UIntFormatter):
    def __init__(self, min_val=0, max_val=2**16 - 1, **kwargs):
        super(UInt16Formatter, self).__init__(min_val, max_val, **kwargs)


class UInt24Formatter(UIntFormatter):
    def __init__(self, min_val=0, max_val=2**24 - 1, **kwargs):
        super(UInt24Formatter, self).__init__(min_val, max_val, **kwargs)


class UInt32Formatter(UIntFormatter):
    def __init__(self, min_val=0, max_val=2**32 - 1, **kwargs):
        super(UInt32Formatter, self).__init__(min_val, max_val, **kwargs)


class FloatFormatter(Formatter):
    """Signed or unsigned floating-point number."""
    def __init__(self, min_val=None, max_val=None, **kwargs):
        super(FloatFormatter, self).__init__(min_val, max_val, **kwargs)

        self.max_val = max_val
        self.min_val = min_val

    def validate(self, str_value):
        try:
            v = self.coerce(str_value)
            if self.max_val is not None and v > self.max_val:
                return False
            if self.min_val is not None and v < self.min_val:
                return False
            return True
        except:
            return False

    def coerce(self, str_value):
        if str_value:
            return float(str_value)
        return str_value


class DoubleFormatter(FloatFormatter):
    pass


class UFloatFormatter(FloatFormatter):
    """Unsigned floating-point number."""
    def __init__(self, min_val=None, max_val=None, **kwargs):
        super(UFloatFormatter, self).__init__(min_val, max_val, **kwargs)
        self.min_val = max(0, min_val)


class UDoubleFormatter(UFloatFormatter):
    pass


class TextFormatter(six.with_metaclass(FormatterMeta, Formatter)):
    pass


class PathFormatter(TextFormatter):
    ERROR_INVALID_NAME = 123

    def __init__(self, exist=False, types=None, **kwargs):
        super(PathFormatter, self).__init__(**kwargs)
        self.exist = exist
        self.types = types

    def validate(self, str_value):
        if self.exist:
            if self.types == 'folder':
                return os.path.isdir(str_value)
            elif self.types == 'file':
                return os.path.isfile(str_value)
            else:
                return os.path.exists(str_value)
        else:
            return self.is_path_exists_or_creatable(str_value)

    def is_pathname_valid(self, pathname):
        '''
        `True` if the passed pathname is a valid pathname for the current OS;
        `False` otherwise.
        '''
        #https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta/9532586
        # If this pathname is either not a string or is but is empty, this pathname
        # is invalid.
        try:
            if not isinstance(pathname, six.string_types) or not pathname:
                return False

            # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
            # if any. Since Windows prohibits path components from containing `:`
            # characters, failing to strip this `:`-suffixed prefix would
            # erroneously invalidate all valid absolute Windows pathnames.
            _, pathname = os.path.splitdrive(pathname)

            # Directory guaranteed to exist. If the current OS is Windows, this is
            # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
            # environment variable); else, the typical root directory.
            root_dirname = os.environ.get('HOMEDRIVE', 'C:') \
                if sys.platform == 'win32' else os.path.sep
            assert os.path.isdir(
                root_dirname)  # ...Murphy and her ironclad Law

            # Append a path separator to this directory if needed.
            root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

            # Test whether each path component split from this pathname is valid or
            # not, ignoring non-existent and non-readable path components.
            for pathname_part in pathname.split(os.path.sep):
                try:
                    os.lstat(root_dirname + pathname_part)
                # If an OS-specific exception is raised, its error code
                # indicates whether this pathname is valid or not. Unless this
                # is the case, this exception implies an ignorable kernel or
                # filesystem complaint (e.g., path not found or inaccessible).
                #
                # Only the following exceptions indicate invalid pathnames:
                #
                # * Instances of the Windows-specific "WindowsError" class
                #   defining the "winerror" attribute whose value is
                #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
                #   fine-grained and hence useful than the generic "errno"
                #   attribute. When a too-long pathname is passed, for example,
                #   "errno" is "ENOENT" (i.e., no such file or directory) rather
                #   than "ENAMETOOLONG" (i.e., file name too long).
                # * Instances of the cross-platform "OSError" class defining the
                #   generic "errno" attribute whose value is either:
                #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
                #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
                except OSError as exc:
                    if hasattr(exc, 'winerror'):
                        if exc.winerror == self.ERROR_INVALID_NAME:
                            return False
                    elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                        return False
        # If a "TypeError" exception was raised, it almost certainly has the
        # error message "embedded NUL character" indicating an invalid pathname.
        except TypeError as exc:
            return False
        # If no exception was raised, all path components and hence this
        # pathname itself are valid. (Praise be to the curmudgeonly python.)
        else:
            return True
        # If any other exception was raised, this is an unrelated fatal issue
        # (e.g., a bug). Permit this exception to unwind the call stack.
        #
        # Did we mention this should be shipped with Python already?

    def is_path_creatable(self, pathname):
        '''
        `True` if the current user has sufficient permissions to create the passed
        pathname; `False` otherwise.
        '''
        # Parent directory of the passed path. If empty, we substitute the current
        # working directory (CWD) instead.
        dirname = os.path.dirname(pathname) or os.getcwd()
        return os.access(dirname, os.W_OK)

    def is_path_exists_or_creatable(self, pathname):
        '''
        `True` if the passed pathname is a valid pathname for the current OS _and_
        either currently exists or is hypothetically creatable; `False` otherwise.

        This function is guaranteed to _never_ raise exceptions.
        '''
        try:
            # To prevent "os" module calls from raising undesirable exceptions on
            # invalid pathnames, is_pathname_valid() is explicitly called first.
            return self.is_pathname_valid(pathname) and (
                os.path.exists(pathname) or self.is_path_creatable(pathname))
        # Report failure on non-fatal filesystem complaints (e.g., connection
        # timeouts, permissions issues) implying this path to be inaccessible. All
        # other exceptions are unrelated fatal issues and should not be caught here.
        except OSError:
            return False


class TimeElapsedFormatter(six.with_metaclass(FormatterMeta, Formatter)):
    """Elapsed time string (HH:MM:SS)."""
    re_validation = '^([1][0-2]|[0]?[0-9]):[0-5][0-9](:[0-5][0-9])?$'


class DateFormatter(Formatter):
    """
    Date string (YYYY-MM-DD).

    Storage format:      wx.DateTime
    Presentation format: YYYY-MM-DD

    Accepts only YYYY-MM-DD format (allows variant separators '/' and '.').
    Accepts dates in range (1000-2999)-(01-12)-(01-31).
    Leading zeros optional in month and day.
    Does not enforce # of days in month.
    """
    def __init__(self, fmt='%Y-%m-%d', **kwargs):
        super(DateFormatter, self).__init__(**kwargs)
        self.str_format = fmt

    def validate(self, str_value):
        value = wx.DateTime()
        return value.ParseDate(str_value) != -1

    def format(self, value):
        return value.Format(self.str_format)

    def coerce(self, str_value):
        """Convert alternate date separators to '-'."""
        value = wx.DateTime()
        value.ParseDate(str_value)
        return value


class DateFormatterMDY(DateFormatter):
    """Alternate date string (MM-DD-YYYY).

    Presentation format: MM-DD-YYYY

    Accepts only MM-DD-YYYY format  (allows variant separators '/' and '.').
    Accepts dates in range (01-12)-(01-31)-(1000-2999).
    Leading zeros optional in month and day.
    Does not enforce # of days in month.
    """
    def __init__(self, fmt='%m-%d-%Y', **kwargs):
        super(DateFormatterMDY, self).__init__(fmt, **kwargs)


class TimeFormatter(DateFormatter):
    """
    Time string (12-hour or 24-hour format, with or without seconds or am/pm).

    Storage format:      HH:MM:SS -- 24-hour format.
    Presentation format: HH:MM    -- 24-hour format.

    Accepts 12-hour or 24-hour format, with or without seconds or am/pm.
    """
    def __init__(self, fmt='%H:%M:%S', **kwargs):
        super(TimeFormatter, self).__init__(fmt, **kwargs)

    def validate(self, str_value):
        value = wx.DateTime()
        return value.ParseTime(str_value) != -1

    def coerce(self, str_value):
        """Convert alternate date separators to '-'."""
        value = wx.DateTime()
        value.ParseTime(str_value)
        return value


class TimeFormatter12H(TimeFormatter):
    """
    Alternate time string (12-hour format, without seconds).

    Storage format:      HH:MM(:SS) -- 24-hour format.
    Presentation format: HH:MM(aa) -- 12-hour format.
                         (aa may be 'am' or 'pm')
    """
    def __init__(self, fmt='%I:%M:%S %p', **kwargs):
        super(TimeFormatter12H, self).__init__(fmt, **kwargs)


class DateTimeFormatter(DateFormatter):
    """
    Date/time string.

    Uses a DateFormatter and a TimeFormatter.
    """

    # Storage format: YYYY-MM-DD HH:MM:SS -- 24-hour format.
    # Presentation format: same as storage format.
    def __init__(self, fmt='%Y-%m-%d %H:%M:%S', **kwargs):
        super(DateTimeFormatter, self).__init__(fmt, **kwargs)

    def validate(self, str_value):
        value = wx.DateTime()
        return value.ParseDateTime(str_value) != -1

    def coerce(self, str_value):
        """Convert alternate date separators to '-'."""
        value = wx.DateTime()
        value.ParseDateTime(str_value)
        return value


class ColorFormatter(Formatter):
    # Storage formate: wx.Colour
    # Presentation format HTML-like syntax: #xxxxxx.
    def validate(self, str_value):
        clr = wx.Colour()
        return clr.Set(str_value)

    def format(self, value):
        if isinstance(value, six.string_types):
            return value
        elif isinstance(value, wx.Colour):
            return value.GetAsString(wx.C2S_HTML_SYNTAX)
        return ""

    def coerce(self, str_value):
        try:
            clr = wx.Colour()
            clr.Set(str_value)
            return clr.GetAsString(wx.C2S_HTML_SYNTAX)
        except:
            return ""


class FontFormatter(Formatter):
    # Storage formate: wx.Font
    def validate(self, str_value):
        font = wx.Font(str_value)
        return font.IsOk()

    def format(self, value):
        if isinstance(value, six.string_types):
            return value
        elif isinstance(value, wx.Font):
            return value.GetNativeFontInfoDesc()
        return ""

    def coerce(self, str_value):
        font = wx.Font(str_value)
        return font


if __name__ == '__main__':

    from .enumtype import EnumType

    def test(tests, fmt, msg):
        print(msg)
        failed, passed = 0, 0
        for t, r in tests:
            actual = fmt.validate(t)
            if actual == r:
                passed += 1
            else:
                print('failed test: {0}, expect {1}, returned {2}'.format(
                    t, r, actual))
                failed += 1
        print('\tPASS: %d tests' % passed)
        print('\tFAIL: %d tests' % failed)

    # ========== EnumFormatter ==========
    print('EnumFormatter')
    TestEnum = EnumType('A', 'B', 'C', D=3)
    formatter = EnumFormatter(TestEnum)
    passCount = 0
    failCount = 0

    # formatter.validValues
    expectedValidValues = [(0, 'A'), (1, 'B'), (2, 'C'), (3, 'D')]
    temp = formatter.validValues()
    if temp == expectedValidValues:
        passCount += 1
    else:
        failCount += 1
        print('formatter.validValues() = %r -- WRONG -- expected %r' %
              (temp, expectedValidValues))

    # formatter.format and formatter.coerce
    for id, name in TestEnum.items():
        # formatter.format
        temp = formatter.format(id)
        if temp == name:
            passCount += 1
        else:
            failCount += 1
            print('formatter.format(%d) = %r -- WRONG -- expected %r' %
                  (id, temp, name))
        # formatter.coerce
        temp = formatter.coerce(name)
        if temp == id:
            passCount += 1
        else:
            failCount += 1
            print('formatter.coerce(%r) = %r -- WRONG -- expected %r' %
                  (name, temp, id))
    if failCount == 0:
        print('\tPASS (%d tests)' % passCount)

    tests = [
        ('a@b.com', True),
        ('a@b', False),
        ('@b', False),
        ('@b.c', False),
        ('@b.us', False),
        ('a@b.us', True),
        ('~joe_public-73@bozo.net', True),
    ]
    test(tests, EmailFormatter(), 'EmailFormatter')

    tests = [('10', True), ('0', True), ('100', True), ('-1', False)]
    test(tests, IntFormatter(0, 100), 'IntFormatter')
    test(tests, UIntFormatter(-1, 100), 'UIntFormatter')
    test(tests, FloatFormatter(0, 100), 'FloatFormatter')
    test(tests, UFloatFormatter(0, 100), 'UFloatFormatter')

    tests = [('0x0a', True), ('0x00', True), ('0x64', True), ('-0x1', False)]
    test(tests, HexFormatter(0, 100), 'HexFormatter')

    tests = [('0b1010', True), ('0b0', True), ('0b1100100', True),
             ('-0b1', False)]
    test(tests, BinFormatter(0, 100), 'BinFormatter')

    tests = [
        ('10', True),
        ('0', True),
        ('127', True),
        ('128', False),
        ('-1', True),
        ('-128', True),
        ('-129', False),
    ]
    test(tests, Int8Formatter(), 'Int8Formatter')
    tests = [
        ('10', True),
        ('0', True),
        ('127', True),
        ('255', True),
        ('256', False),
        ('-1', False),
    ]
    test(tests, UInt8Formatter(), 'UInt8Formatter')



# -*- coding: utf-8 -*-
#
# Sorted dict implementation.

from .sortedset import SortedSet
from .sortedlist import SortedList, recursive_repr, SortedListWithKey
from collections import Set, Sequence
from collections import KeysView as AbstractKeysView
from collections import ValuesView as AbstractValuesView
from collections import ItemsView as AbstractItemsView

from functools import wraps
from sys import hexversion

_NotGiven = object()

def not26(func):
    """Function decorator for methods not implemented in Python 2.6."""

    @wraps(func)
    def errfunc(*args, **kwargs):
        raise NotImplementedError

    if hexversion < 0x02070000:
        return errfunc
    else:
        return func

class _IlocWrapper:
    def __init__(self, _dict):
        self._dict = _dict
    def __len__(self):
        return len(self._dict)
    def __getitem__(self, index):
        """
        Very efficiently return the key at index *index* in iteration. Supports
        negative indices and slice notation. Raises IndexError on invalid
        *index*.
        """
        return self._dict._list[index]
    def __delitem__(self, index):
        """
        Remove the ``sdict[sdict.iloc[index]]`` from *sdict*. Supports negative
        indices and slice notation. Raises IndexError on invalid *index*.
        """
        _temp = self._dict
        _list = _temp._list
        _delitem = _temp._delitem

        if isinstance(index, slice):
            keys = _list[index]
            del _list[index]
            for key in keys:
                _delitem(key)
        else:
            key = _list[index]
            del _list[index]
            _delitem(key)

class SortedDict(dict):
    """
    A SortedDict provides the same methods as a dict.  Additionally, a
    SortedDict efficiently maintains its keys in sorted order. Consequently, the
    keys method will return the keys in sorted order, the popitem method will
    remove the item with the highest key, etc.
    """
    def __init__(self, *args, **kwargs):
        """
        A SortedDict provides the same methods as a dict.  Additionally, a
        SortedDict efficiently maintains its keys in sorted order. Consequently,
        the keys method will return the keys in sorted order, the popitem method
        will remove the item with the highest key, etc.

        An optional *key* argument defines a callable that, like the `key`
        argument to Python's `sorted` function, extracts a comparison key from
        each dict key. If no function is specified, the default compares the
        dict keys directly. The `key` argument must be provided as a positional
        argument and must come before all other arguments.

        An optional *load* argument defines the load factor of the internal list
        used to maintain sort order. If present, this argument must come before
        an iterable. The default load factor of '1000' works well for lists from
        tens to tens of millions of elements.  Good practice is to use a value
        that is the cube root of the list size.  With billions of elements, the
        best load factor depends on your usage.  It's best to leave the load
        factor at the default until you start benchmarking.

        An optional *iterable* argument provides an initial series of items to
        populate the SortedDict.  Each item in the series must itself contain
        two items.  The first is used as a key in the new dictionary, and the
        second as the key's value. If a given key is seen more than once, the
        last value associated with it is retained in the new dictionary.

        If keyword arguments are given, the keywords themselves with their
        associated values are added as items to the dictionary. If a key is
        specified both in the positional argument and as a keyword argument, the
        value associated with the keyword is retained in the dictionary. For
        example, these all return a dictionary equal to ``{"one": 2, "two":
        3}``:

        * ``SortedDict(one=2, two=3)``
        * ``SortedDict({'one': 2, 'two': 3})``
        * ``SortedDict(zip(('one', 'two'), (2, 3)))``
        * ``SortedDict([['two', 3], ['one', 2]])``

        The first example only works for keys that are valid Python
        identifiers; the others work with any valid keys.
        """
        if len(args) > 0 and (args[0] is None or callable(args[0])):
            self._key = args[0]
            args = args[1:]
        else:
            self._key = None

        if len(args) > 0 and type(args[0]) == int:
            self._load = args[0]
            args = args[1:]
        else:
            self._load = 1000

        if self._key is None:
            self._list = SortedList(load=self._load)
        else:
            self._list = SortedListWithKey(key=self._key, load=self._load)

        # Cache function pointers to dict methods.

        _dict = super(SortedDict, self)
        self._dict = _dict
        self._clear = _dict.clear
        self._delitem = _dict.__delitem__
        self._iter = _dict.__iter__
        self._pop = _dict.pop
        self._setdefault = _dict.setdefault
        self._setitem = _dict.__setitem__
        self._dict_update = _dict.update

        # Cache function pointers to SortedList methods.

        _list = self._list
        self._list_add = _list.add
        self.bisect_left = _list.bisect_left
        self.bisect = _list.bisect_right
        self.bisect_right = _list.bisect_right
        self._list_clear = _list.clear
        self.index = _list.index
        self._list_pop = _list.pop
        self._list_remove = _list.remove
        self._list_update = _list.update
        self.irange = _list.irange
        self.islice = _list.islice

        if self._key is not None:
            self.bisect_key_left = _list.bisect_key_left
            self.bisect_key_right = _list.bisect_key_right
            self.bisect_key = _list.bisect_key
            self.irange_key = _list.irange_key

        self.iloc = _IlocWrapper(self)

        self._update(*args, **kwargs)

    def clear(self):
        """Remove all elements from the dictionary."""
        self._clear()
        self._list_clear()

    def __delitem__(self, key):
        """
        Remove ``d[key]`` from *d*.  Raises a KeyError if *key* is not in the
        dictionary.
        """
        self._delitem(key)
        self._list_remove(key)

    def __iter__(self):
        """
        Return an iterator over the sorted keys of the dictionary.

        Iterating the Mapping while adding or deleting keys may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return iter(self._list)

    def __reversed__(self):
        """
        Return a reversed iterator over the sorted keys of the dictionary.

        Iterating the Mapping while adding or deleting keys may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return reversed(self._list)

    def __setitem__(self, key, value):
        """Set `d[key]` to *value*."""
        if key not in self:
            self._list_add(key)
        self._setitem(key, value)

    def copy(self):
        """Return a shallow copy of the sorted dictionary."""
        return self.__class__(self._key, self._load, self._iteritems())

    __copy__ = copy

    @classmethod
    def fromkeys(cls, seq, value=None):
        """
        Create a new dictionary with keys from *seq* and values set to *value*.
        """
        return cls((key, value) for key in seq)

    if hexversion < 0x03000000:
        def items(self):
            """
            Return a list of the dictionary's items (``(key, value)`` pairs).
            """
            return list(self._iteritems())
    else:
        def items(self):
            """
            Return a new ItemsView of the dictionary's items.  In addition to
            the methods provided by the built-in `view` the ItemsView is
            indexable (e.g. ``d.items()[5]``).
            """
            return ItemsView(self)

    def iteritems(self):
        """
        Return an iterator over the items (``(key, value)`` pairs).

        Iterating the Mapping while adding or deleting keys may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return iter((key, self[key]) for key in self._list)

    _iteritems = iteritems

    if hexversion < 0x03000000:
        def keys(self):
            """Return a SortedSet of the dictionary's keys."""
            return SortedSet(self._list, key=self._key, load=self._load)
    else:
        def keys(self):
            """
            Return a new KeysView of the dictionary's keys.  In addition to the
            methods provided by the built-in `view` the KeysView is indexable
            (e.g. ``d.keys()[5]``).
            """
            return KeysView(self)

    def iterkeys(self):
        """
        Return an iterator over the sorted keys of the Mapping.

        Iterating the Mapping while adding or deleting keys may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return iter(self._list)

    if hexversion < 0x03000000:
        def values(self):
            """Return a list of the dictionary's values."""
            return list(self._itervalues())
    else:
        def values(self):
            """
            Return a new :class:`ValuesView` of the dictionary's values.
            In addition to the methods provided by the built-in `view` the
            ValuesView is indexable (e.g., ``d.values()[5]``).
            """
            return ValuesView(self)

    def itervalues(self):
        """
        Return an iterator over the values of the Mapping.

        Iterating the Mapping while adding or deleting keys may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return iter(self[key] for key in self._list)

    _itervalues = itervalues

    def pop(self, key, default=_NotGiven):
        """
        If *key* is in the dictionary, remove it and return its value,
        else return *default*. If *default* is not given and *key* is not in
        the dictionary, a KeyError is raised.
        """
        if key in self:
            self._list_remove(key)
            return self._pop(key)
        else:
            if default is _NotGiven:
                raise KeyError(key)
            else:
                return default

    def popitem(self, last=True):
        """
        Remove and return a ``(key, value)`` pair from the dictionary. If
        last=True (default) then remove the *greatest* `key` from the
        diciontary. Else, remove the *least* key from the dictionary.

        If the dictionary is empty, calling `popitem` raises a
        KeyError`.
        """
        if not len(self):
            raise KeyError('popitem(): dictionary is empty')

        key = self._list_pop(-1 if last else 0)
        value = self._pop(key)

        return (key, value)

    def setdefault(self, key, default=None):
        """
        If *key* is in the dictionary, return its value.  If not, insert *key*
        with a value of *default* and return *default*.  *default* defaults to
        ``None``.
        """
        if key in self:
            return self[key]
        else:
            self._setitem(key, default)
            self._list_add(key)
            return default

    def update(self, *args, **kwargs):
        """
        Update the dictionary with the key/value pairs from *other*, overwriting
        existing keys.

        *update* accepts either another dictionary object or an iterable of
        key/value pairs (as a tuple or other iterable of length two).  If
        keyword arguments are specified, the dictionary is then updated with
        those key/value pairs: ``d.update(red=1, blue=2)``.
        """
        if not len(self):
            self._dict_update(*args, **kwargs)
            self._list_update(self._iter())
            return

        if (len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], dict)):
            pairs = args[0]
        else:
            pairs = dict(*args, **kwargs)

        if (10 * len(pairs)) > len(self):
            self._dict_update(pairs)
            self._list_clear()
            self._list_update(self._iter())
        else:
            for key in pairs:
                self[key] = pairs[key]

    _update = update

    @not26
    def viewkeys(self):
        """
        In Python 2.7 and later, return a new `KeysView` of the dictionary's
        keys.

        In Python 2.6, raise a NotImplementedError.
        """
        return KeysView(self)

    @not26
    def viewvalues(self):
        """
        In Python 2.7 and later, return a new `ValuesView` of the dictionary's
        values.

        In Python 2.6, raise a NotImplementedError.
        """
        return ValuesView(self)

    @not26
    def viewitems(self):
        """
        In Python 2.7 and later, return a new `ItemsView` of the dictionary's
        items.

        In Python 2.6, raise a NotImplementedError.
        """
        return ItemsView(self)

    def __reduce__(self):
        return (self.__class__, (self._key, self._load, list(self._iteritems())))

    @recursive_repr
    def __repr__(self):
        temp = '{0}({1}, {2}, {{{3}}})'
        items = ', '.join('{0}: {1}'.format(repr(key), repr(self[key]))
                          for key in self._list)
        return temp.format(
            self.__class__.__name__,
            repr(self._key),
            repr(self._load),
            items
        )

    def _check(self):
        self._list._check()
        assert len(self) == len(self._list)
        assert all(val in self for val in self._list)

class KeysView(AbstractKeysView, Set, Sequence):
    """
    A KeysView object is a dynamic view of the dictionary's keys, which
    means that when the dictionary's keys change, the view reflects
    those changes.

    The KeysView class implements the Set and Sequence Abstract Base Classes.
    """
    if hexversion < 0x03000000:
        def __init__(self, sorted_dict):
            """
            Initialize a KeysView from a SortedDict container as *sorted_dict*.
            """
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.viewkeys()
    else:
        def __init__(self, sorted_dict):
            """
            Initialize a KeysView from a SortedDict container as *sorted_dict*.
            """
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.keys()
    def __len__(self):
        """Return the number of entries in the dictionary."""
        return len(self._view)
    def __contains__(self, key):
        """
        Return True if and only if *key* is one of the underlying dictionary's
        keys.
        """
        return key in self._view
    def __iter__(self):
        """
        Return an iterable over the keys in the dictionary. Keys are iterated
        over in their sorted order.

        Iterating views while adding or deleting entries in the dictionary may
        raise a `RuntimeError` or fail to iterate over all entries.
        """
        return iter(self._list)
    def __getitem__(self, index):
        """Return the key at position *index*."""
        return self._list[index]
    def __reversed__(self):
        """
        Return a reversed iterable over the keys in the dictionary. Keys are
        iterated over in their reverse sort order.

        Iterating views while adding or deleting entries in the dictionary may
        raise a RuntimeError or fail to iterate over all entries.
        """
        return reversed(self._list)
    def index(self, value, start=None, stop=None):
        """
        Return the smallest *k* such that `keysview[k] == value` and `start <= k
        < end`.  Raises `KeyError` if *value* is not present.  *stop* defaults
        to the end of the set.  *start* defaults to the beginning.  Negative
        indexes are supported, as for slice indices.
        """
        return self._list.index(value, start, stop)
    def count(self, value):
        """Return the number of occurrences of *value* in the set."""
        return 1 if value in self._view else 0
    def __eq__(self, that):
        """Test set-like equality with *that*."""
        return self._view == that
    def __ne__(self, that):
        """Test set-like inequality with *that*."""
        return self._view != that
    def __lt__(self, that):
        """Test whether self is a proper subset of *that*."""
        return self._view < that
    def __gt__(self, that):
        """Test whether self is a proper superset of *that*."""
        return self._view > that
    def __le__(self, that):
        """Test whether self is contained within *that*."""
        return self._view <= that
    def __ge__(self, that):
        """Test whether *that* is contained within self."""
        return self._view >= that
    def __and__(self, that):
        """Return a SortedSet of the intersection of self and *that*."""
        return SortedSet(self._view & that)
    def __or__(self, that):
        """Return a SortedSet of the union of self and *that*."""
        return SortedSet(self._view | that)
    def __sub__(self, that):
        """Return a SortedSet of the difference of self and *that*."""
        return SortedSet(self._view - that)
    def __xor__(self, that):
        """Return a SortedSet of the symmetric difference of self and *that*."""
        return SortedSet(self._view ^ that)
    if hexversion < 0x03000000:
        def isdisjoint(self, that):
            """Return True if and only if *that* is disjoint with self."""
            return not any(key in self._list for key in that)
    else:
        def isdisjoint(self, that):
            """Return True if and only if *that* is disjoint with self."""
            return self._view.isdisjoint(that)
    @recursive_repr
    def __repr__(self):
        return 'SortedDict_keys({0})'.format(repr(list(self)))

class ValuesView(AbstractValuesView, Sequence):
    """
    A ValuesView object is a dynamic view of the dictionary's values, which
    means that when the dictionary's values change, the view reflects those
    changes.

    The ValuesView class implements the Sequence Abstract Base Class.
    """
    if hexversion < 0x03000000:
        def __init__(self, sorted_dict):
            """
            Initialize a ValuesView from a SortedDict container as
            *sorted_dict*.
            """
            self._dict = sorted_dict
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.viewvalues()
    else:
        def __init__(self, sorted_dict):
            """
            Initialize a ValuesView from a SortedDict container as
            *sorted_dict*.
            """
            self._dict = sorted_dict
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.values()
    def __len__(self):
        """Return the number of entries in the dictionary."""
        return len(self._dict)
    def __contains__(self, value):
        """
        Return True if and only if *value* is in the underlying Mapping's
        values.
        """
        return value in self._view
    def __iter__(self):
        """
        Return an iterator over the values in the dictionary.  Values are
        iterated over in sorted order of the keys.

        Iterating views while adding or deleting entries in the dictionary may
        raise a `RuntimeError` or fail to iterate over all entries.
        """
        _dict = self._dict
        return iter(_dict[key] for key in self._list)
    def __getitem__(self, index):
        """
        Efficiently return value at *index* in iteration.

        Supports slice notation and negative indexes.
        """
        _dict, _list = self._dict, self._list
        if isinstance(index, slice):
            return [_dict[key] for key in _list[index]]
        else:
            return _dict[_list[index]]
    def __reversed__(self):
        """
        Return a reverse iterator over the values in the dictionary.  Values are
        iterated over in reverse sort order of the keys.

        Iterating views while adding or deleting entries in the dictionary may
        raise a `RuntimeError` or fail to iterate over all entries.
        """
        _dict = self._dict
        return iter(_dict[key] for key in reversed(self._list))
    def index(self, value):
        """
        Return index of *value* in self.

        Raises ValueError if *value* is not found.
        """
        for idx, val in enumerate(self):
            if value == val:
                return idx
        else:
            raise ValueError('{0} is not in dict'.format(repr(value)))
    if hexversion < 0x03000000:
        def count(self, value):
            """Return the number of occurrences of *value* in self."""
            return sum(1 for val in self._dict.itervalues() if val == value)
    else:
        def count(self, value):
            """Return the number of occurrences of *value* in self."""
            return sum(1 for val in self._dict.values() if val == value)
    def __lt__(self, that):
        raise TypeError
    def __gt__(self, that):
        raise TypeError
    def __le__(self, that):
        raise TypeError
    def __ge__(self, that):
        raise TypeError
    def __and__(self, that):
        raise TypeError
    def __or__(self, that):
        raise TypeError
    def __sub__(self, that):
        raise TypeError
    def __xor__(self, that):
        raise TypeError
    @recursive_repr
    def __repr__(self):
        return 'SortedDict_values({0})'.format(repr(list(self)))

class ItemsView(AbstractItemsView, Set, Sequence):
    """
    An ItemsView object is a dynamic view of the dictionary's ``(key,
    value)`` pairs, which means that when the dictionary changes, the
    view reflects those changes.

    The ItemsView class implements the Set and Sequence Abstract Base Classes.
    However, the set-like operations (``&``, ``|``, ``-``, ``^``) will only
    operate correctly if all of the dictionary's values are hashable.
    """
    if hexversion < 0x03000000:
        def __init__(self, sorted_dict):
            """
            Initialize an ItemsView from a SortedDict container as
            *sorted_dict*.
            """
            self._dict = sorted_dict
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.viewitems()
    else:
        def __init__(self, sorted_dict):
            """
            Initialize an ItemsView from a SortedDict container as
            *sorted_dict*.
            """
            self._dict = sorted_dict
            self._list = sorted_dict._list
            self._view = sorted_dict._dict.items()
    def __len__(self):
        """Return the number of entries in the dictionary."""
        return len(self._view)
    def __contains__(self, key):
        """
        Return True if and only if *key* is one of the underlying dictionary's
        items.
        """
        return key in self._view
    def __iter__(self):
        """
        Return an iterable over the items in the dictionary. Items are iterated
        over in their sorted order.

        Iterating views while adding or deleting entries in the dictionary may
        raise a `RuntimeError` or fail to iterate over all entries.
        """
        _dict = self._dict
        return iter((key, _dict[key]) for key in self._list)
    def __getitem__(self, index):
        """Return the item as position *index*."""
        _dict, _list = self._dict, self._list
        if isinstance(index, slice):
            return [(key, _dict[key]) for key in _list[index]]
        else:
            key = _list[index]
            return (key, _dict[key])
    def __reversed__(self):
        """
        Return a reversed iterable over the items in the dictionary. Items are
        iterated over in their reverse sort order.

        Iterating views while adding or deleting entries in the dictionary may
        raise a RuntimeError or fail to iterate over all entries.
        """
        _dict = self._dict
        return iter((key, _dict[key]) for key in reversed(self._list))
    def index(self, key, start=None, stop=None):
        """
        Return the smallest *k* such that `itemssview[k] == key` and `start <= k
        < end`.  Raises `KeyError` if *key* is not present.  *stop* defaults
        to the end of the set.  *start* defaults to the beginning.  Negative
        indexes are supported, as for slice indices.
        """
        temp, value = key
        pos = self._list.index(temp, start, stop)
        if value == self._dict[temp]:
            return pos
        else:
            raise ValueError('{0} is not in dict'.format(repr(key)))
    def count(self, item):
        """Return the number of occurrences of *item* in the set."""
        key, value = item
        return 1 if key in self._dict and self._dict[key] == value else 0
    def __eq__(self, that):
        """Test set-like equality with *that*."""
        return self._view == that
    def __ne__(self, that):
        """Test set-like inequality with *that*."""
        return self._view != that
    def __lt__(self, that):
        """Test whether self is a proper subset of *that*."""
        return self._view < that
    def __gt__(self, that):
        """Test whether self is a proper superset of *that*."""
        return self._view > that
    def __le__(self, that):
        """Test whether self is contained within *that*."""
        return self._view <= that
    def __ge__(self, that):
        """Test whether *that* is contained within self."""
        return self._view >= that
    def __and__(self, that):
        """Return a SortedSet of the intersection of self and *that*."""
        return SortedSet(self._view & that)
    def __or__(self, that):
        """Return a SortedSet of the union of self and *that*."""
        return SortedSet(self._view | that)
    def __sub__(self, that):
        """Return a SortedSet of the difference of self and *that*."""
        return SortedSet(self._view - that)
    def __xor__(self, that):
        """Return a SortedSet of the symmetric difference of self and *that*."""
        return SortedSet(self._view ^ that)
    if hexversion < 0x03000000:
        def isdisjoint(self, that):
            """Return True if and only if *that* is disjoint with self."""
            _dict = self._dict
            for key, value in that:
                if key in _dict and _dict[key] == value:
                    return False
            return True
    else:
        def isdisjoint(self, that):
            """Return True if and only if *that* is disjoint with self."""
            return self._view.isdisjoint(that)
    @recursive_repr
    def __repr__(self):
        return 'SortedDict_items({0})'.format(repr(list(self)))



import os
import sys
import re
import traceback
import subprocess as sp
import keyword
import time
import pydoc
import six.moves.builtins as __builtin__
import six
import wx
import wx.py.shell as pyshell
import wx.py.dispatcher as dp
from wx import stc
from wx.py.pseudo import PseudoFile
from .debugger import EngineDebugger
from ..version import __version__


# in linux, the multiprocessing/process.py/_bootstrap will call
# sys.stdin.close(), which is missing in wx.py.pseudo.PseudoFile
def PseudoFile_close(self):
    pass


PseudoFile.close = PseudoFile_close

aliasDict = {}


def magicSingle(command):
    if command == '':  # Pass if command is blank
        return command

    first_space = command.find(' ')

    if command[0] == ' ':  # Pass if command begins with a space
        pass
    elif command[0] == '?':  # Do help if starts with ?
        command = 'help(' + command[1:] + ')'
    elif command[0] == '!':  # Use os.system if starts with !
        command = 'sx("' + command[1:] + '")'
    elif command in ('ls', 'pwd'):
        # automatically use ls and pwd with no arguments
        command = command + '()'
    elif command[:3] in ('ls ', 'cd '):
        # when using the 'ls ' or 'cd ' constructs, fill in both parentheses and quotes
        command = command[:2] + '("' + command[3:] + '")'
    elif command[:5] in ('help ', ):
        command = command[:4] + '("' + command[5:] + '")'
    elif command[:6] == 'close ':
        arg = command[6:]
        if arg.strip() == 'all':
            # when using the close', fill in both parentheses and quotes
            command = command[:5] + '("' + command[6:] + '")'
    elif command[:5] == 'clear':
        command = command[:5] + '()'
    elif command[:5] == 'alias':
        c = command[5:].lstrip().split(' ')
        if len(c) < 2:
            # delete the alias if exists
            if len(c) == 1:
                aliasDict.pop(c[0], None)
            command = ''
        else:
            n, v = c[0], ' '.join(c[1:])
            aliasDict[n] = v
            command = ''
    elif command.split(' ')[0] in six.iterkeys(aliasDict):
        c = command.split(' ')
        if len(c) < 2:
            command = 'sx("' + aliasDict[c[0]] + '")'
        else:
            command = 'sx("' + aliasDict[c[0]] + ' ' + ' '.join(c[1:]) + '")'
    elif first_space != -1:
        # if there is at least one space, add parentheses at beginning and end
        cmds = command.split(' ')
        if len(cmds) > 1:
            wd1 = cmds[0]
            wd2 = cmds[1]
            i = 1
            while wd2 == '':
                i += 1
                if len(cmds) == i:
                    break
                wd2 = cmds[i]
            if wd2 == '':
                return command
            if (wd1[0].isalpha() or wd1[0] == '_') and (wd2[0].isalnum() or\
                    (wd2[0] in """."'_""")) and \
                    not keyword.iskeyword(wd1) and not keyword.iskeyword(wd2):
                if wd1.replace('.', '').replace('_', '').isalnum():
                    # add parentheses where the first space was and at the end... hooray!
                    command = wd1 + '(' + command[(first_space + 1):] + ')'
    return command


def _help(command):
    try:
        print(pydoc.plain(pydoc.render_doc(str(command), "Help on %s")))
    except:
        print('No help found on "%s"' % command)


def magic(command):
    continuations = wx.py.parse.testForContinuations(command)
    if len(continuations) == 2:  # Error case...
        return command
    elif len(continuations) == 4:
        stringContinuationList, indentationBlockList, \
        lineContinuationList, parentheticalContinuationList = continuations

    commandList = []
    firstLine = True
    for i in command.split('\n'):
        if firstLine:
            commandList.append(magicSingle(i))
        elif stringContinuationList.pop(0) is False and \
              indentationBlockList.pop(0) is False and \
              lineContinuationList.pop(0) is False and \
              parentheticalContinuationList.pop(0) is False:
            commandList.append(magicSingle(
                i))  # unless this is in a larger expression, use magic
        else:
            commandList.append(i)

        firstLine = False

    return '\n'.join(commandList)


def sx(cmd, *args, **kwds):
    wait = True
    # append '&' to capture the output
    if cmd[-1] == '&':
        wait = False
        cmd = cmd[0:-1]
    startupinfo = None
    if wx.Platform == '__WXMSW__':
        startupinfo = sp.STARTUPINFO()
        startupinfo.dwFlags |= sp.STARTF_USESHOWWINDOW
    # try the standalone command first
    try:
        if wait:
            p = sp.Popen(cmd.split(' '),
                         startupinfo=startupinfo,
                         stdout=sp.PIPE,
                         stderr=sp.PIPE)
            dp.send('shell.write_out', text=p.stdout.read().decode())
        else:
            p = sp.Popen(cmd.split(' '), startupinfo=startupinfo)
        return
    except:
        traceback.print_exc(file=sys.stdout)
    # try the shell command
    try:
        if wait:
            p = sp.Popen(cmd.split(' '),
                         startupinfo=startupinfo,
                         shell=True,
                         stdout=sp.PIPE,
                         stderr=sp.PIPE)
            dp.send('shell.write_out', text=p.stdout.read().decode())
        else:
            p = sp.Popen(cmd.split(' '), startupinfo=startupinfo, shell=True)
        return
    except:
        traceback.print_exc(file=sys.stdout)


class Shell(pyshell.Shell):
    ID_COPY_PLUS = wx.NewId()
    ID_PASTE_PLUS = wx.NewId()

    def __init__(self,
                 parent,
                 id=-1,
                 pos=wx.DefaultPosition,
                 size=wx.DefaultSize,
                 style=wx.CLIP_CHILDREN,
                 introText='',
                 locals=None,
                 InterpClass=None,
                 startupScript=None,
                 execStartupScript=True,
                 *args,
                 **kwds):
        # variables used in push, which may be called by
        # wx.py.shell.Shell.__init__ when execStartupScript is True
        self.enable_debugger = False
        self.silent = True
        pyshell.Shell.__init__(self, parent, id, pos, size, style, introText,
                               locals, InterpClass, startupScript,
                               execStartupScript, useStockId=False, *args, **kwds)
        wx.CallAfter(self.redirectStdout, True)
        wx.CallAfter(self.redirectStderr, True)
        #self.redirectStdout(True)
        #self.redirectStderr(True)
        # the default sx function (!cmd to run external command) does not work
        # on windows
        __builtin__.sx = sx
        self.callTipInsert = False
        self.searchHistory = True
        self.silent = False
        self.autoIndent = True
        self.running = False
        self.debugger = EngineDebugger()
        self.LoadHistory()
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnLeftDClick)
        self.Bind(wx.EVT_MENU, self.OnProcessMenu)
        self.interp.locals['clear'] = self.clear
        self.interp.locals['help'] = _help
        self.interp.locals['on'] = True
        self.interp.locals['off'] = False
        dp.connect(self.writeOut, 'shell.write_out')
        dp.connect(self.runCommand, 'shell.run')
        dp.connect(self.debugPrompt, 'shell.prompt')
        dp.connect(self.addHistory, 'shell.add_to_history')
        dp.connect(self.IsDebuggerOn, 'debugger.debugging')
        dp.connect(self.getAutoCompleteList, 'shell.auto_complete_list')
        dp.connect(self.getAutoCompleteKeys, 'shell.auto_complete_keys')
        dp.connect(self.getAutoCallTip, 'shell.auto_call_tip')
        dp.connect(self.OnActivatePanel, 'frame.activate_panel')
        dp.connect(self.OnActivate, 'frame.activate')
        dp.connect(self.OnFrameClosing, 'frame.closing')

    def OnContextMenu(self, evt):
        menu = wx.Menu()
        menu.Append(wx.ID_UNDO, "Undo")
        menu.Append(wx.ID_REDO, "Redo")
        menu.AppendSeparator()
        menu.Append(wx.ID_CUT, "Cut")
        menu.Append(wx.ID_COPY, "Copy")
        menu.Append(self.ID_COPY_PLUS, "Copy with prompt")
        menu.Append(wx.ID_PASTE, "Paste")
        menu.Append(self.ID_PASTE_PLUS, "Paste & run")
        menu.Append(wx.ID_CLEAR, "Clear")
        menu.AppendSeparator()
        menu.Append(wx.ID_SELECTALL, "Select All")
        self.PopupMenu(menu)

    def OnProcessMenu(self, event):
        eid = event.GetId()
        cmd = {
            wx.ID_CUT: self.Cut,
            wx.ID_CLEAR: self.Clear,
            wx.ID_COPY: self.Copy,
            self.ID_COPY_PLUS: self.CopyWithPrompts,
            wx.ID_PASTE: self.Paste,
            self.ID_PASTE_PLUS: self.PasteAndRun,
            wx.ID_UNDO: self.Undo,
            wx.ID_REDO: self.Redo,
            wx.ID_SELECTALL: self.SelectAll
        }
        fun = cmd.get(eid, None)
        if fun:
            fun()

    def Destroy(self):
        self.debugger.release()
        # save command history
        dp.send('frame.set_config', group='shell', history=self.history)
        dp.send('frame.set_config', group='shell', alias=aliasDict)

        dp.disconnect(self.writeOut, 'shell.write_out')
        dp.disconnect(self.runCommand, 'shell.run')
        dp.disconnect(self.debugPrompt, 'shell.prompt')
        dp.disconnect(self.addHistory, 'shell.add_to_history')
        dp.disconnect(self.IsDebuggerOn, 'debugger.debugging')
        dp.disconnect(self.getAutoCompleteList, 'shell.auto_complete_list')
        dp.disconnect(self.getAutoCompleteKeys, 'shell.auto_complete_keys')
        dp.disconnect(self.getAutoCallTip, 'shell.auto_call_tip')
        dp.disconnect(self.OnActivatePanel, 'frame.activate_panel')
        dp.disconnect(self.OnActivate, 'frame.activate')
        dp.disconnect(self.OnFrameClosing, 'frame.closing')
        super(Shell, self).Destroy()

    def OnFrameClosing(self, event):
        """the frame is exiting"""
        if self.IsDebuggerOn() and event.CanVeto():
            # stop closing if the debugger is running, otherwise it may crash
            # as some events (e.g., ID_DEBUG_SCRIPT in editor) may be
            # called after some window (e.g., editor) has been destroyed.
            msg = 'The debugger is running.\n Turn it off first!'
            dlg = wx.MessageDialog(self.GetTopLevelParent(), msg, 'bsmedit',
                                   wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            event.Veto()

    def evaluate(self, word):
        if word in six.iterkeys(self.interp.locals):
            return self.interp.locals[word]
        try:
            self.interp.locals[word] = eval(word, self.interp.locals)
            return self.interp.locals[word]
        except:
            try:
                getattr(self, 'AutoCompleteIgnore').index(word)
                return None
            except:
                try:
                    components = word.split('.')
                    try:
                        mod = __import__(word)
                    except:
                        if len(components) < 2:
                            return None
                        mod = '.'.join(components[:-1])
                        try:
                            mod = __import__(mod)
                        except:
                            return None
                    for comp in components[1:]:
                        mod = getattr(mod, comp)
                    self.interp.locals[word] = mod
                    return mod
                except:
                    return None

    def Paste(self):
        """Replace selection with clipboard contents."""
        if self.CanPaste() and wx.TheClipboard.Open():
            ps2 = str(sys.ps2)
            # on mac 11.4, it always return false
            if True:#wx.TheClipboard.IsSupported(wx.DataFormat(wx.DF_TEXT)):
                data = wx.TextDataObject()
                if wx.TheClipboard.GetData(data):
                    self.ReplaceSelection('')
                    command = data.GetText()
                    command = command.rstrip()
                    command = self.fixLineEndings(command)
                    command = self.lstripPrompt(text=command)
                    command = command.replace(os.linesep + ps2, '\n')
                    command = command.replace(os.linesep, '\n')
                    command = command.replace('\n', os.linesep + ps2)
                    self.write(command)
            wx.TheClipboard.Close()

    def PasteAndRun(self):
        """Replace selection with clipboard contents, run commands."""
        text = ''
        if wx.TheClipboard.Open():
            if True:#wx.TheClipboard.IsSupported(wx.DataFormat(wx.DF_TEXT)):
                data = wx.TextDataObject()
                if wx.TheClipboard.GetData(data):
                    text = data.GetText()
            wx.TheClipboard.Close()
        if text:
            self.Execute(text)

    def getAutoCompleteKeys(self):
        return self.interp.getAutoCompleteKeys()

    def getAutoCompleteList(self, command, *args, **kwds):
        # remove additional key from wx.py.dispatcher.send
        kwds.pop('sender', None)
        kwds.pop('signal', None)
        try:
            cmd = wx.py.introspect.getRoot(command, '.')
            self.evaluate(cmd)
        except:
            pass
        return self.interp.getAutoCompleteList(command, *args, **kwds)

    def getAutoCallTip(self, command, *args, **kwds):
        # remove additional key from wx.py.dispatcher.send
        kwds.pop('sender', None)
        kwds.pop('signal', None)
        return self.interp.getCallTip(command, *args, **kwds)

    def autoCompleteShow(self, command, offset=0):
        try:
            command = command.strip()
            # deal with the case "fun(arg.", which will return "arg."
            for i in range(len(command)):
                c = command[-i - 1]
                if not (c.isalnum() or c in ('_', '.')):
                    command = command[-i:]
                    break
            cmd = wx.py.introspect.getRoot(command, '.')
            self.evaluate(cmd)
        except:
            pass
        super(Shell, self).autoCompleteShow(command, offset)

    def IsDebuggerOn(self):
        """check if the debugger is on"""
        if not self.debugger:
            return False
        return self.enable_debugger

    def SetSelection(self, start, end):
        self.SetSelectionStart(start)
        self.SetSelectionEnd(end)
        if end < start:
            self.SetAnchor(start)

    def LoadHistory(self):
        self.clearHistory()
        resp = dp.send('frame.get_config', group='shell', key='history')
        if resp and resp[0][1]:
            self.history = resp[0][1]
        resp = dp.send('frame.get_config', group='shell', key='alias')
        if resp and resp[0][1]:
            aliasDict.update(resp[0][1])

    def OnKillFocus(self, event):
        if self.CallTipActive():
            self.CallTipCancel()
        event.Skip()

    def OnActivate(self, activate):
        # not work on mac
        if wx.Platform == '__WXMAC__':
            return

        if self.AutoCompActive():
            wx.CallAfter(self.AutoCompCancel)

    def OnActivatePanel(self, pane):
        if pane != self:
            if self.AutoCompActive():
                wx.CallAfter(self.AutoCompCancel)

    def OnUpdateUI(self, evt):
        eid = evt.GetId()
        if eid in (wx.ID_CUT, wx.ID_CLEAR):
            evt.Enable(self.CanCut())
        elif eid in (wx.ID_COPY, self.ID_COPY_PLUS):
            evt.Enable(self.CanCopy())
        elif eid in (wx.ID_PASTE, self.ID_PASTE_PLUS):
            evt.Enable(self.CanPaste())
        elif eid == wx.ID_UNDO:
            evt.Enable(self.CanUndo())
        elif eid == wx.ID_REDO:
            evt.Enable(self.CanRedo())
        # update the caret position so that it is always in valid area
        self.UpdateCaretPos()
        super(Shell, self).OnUpdateUI(evt)

    def UpdateCaretPos(self):
        # when editing the command, do not allow moving the caret to
        # readonly area
        if not self.CanEdit() and \
                (self.GetCurrentLine() == self.LineFromPosition(self.promptPosEnd)):
            self.GotoPos(self.promptPosEnd)

    def OnKeyDown(self, event):
        """Key down event handler."""
        key = event.GetKeyCode()
        # If the auto-complete window is up let it do its thing.
        if self.AutoCompActive():
            event.Skip()
            return

        shiftDown = event.ShiftDown()
        controlDown = event.ControlDown()
        rawControlDown = event.RawControlDown()
        altDown = event.AltDown()
        canEdit = self.CanEdit()

        # If it is a letter or digit and the cursor is in readonly section,
        # move the cursor to the end of file
        if not canEdit and (not shiftDown) and (not controlDown) and (not altDown)\
            and (not rawControlDown) and key < 256 and (chr(key).isalnum() or\
               (key == wx.WXK_SPACE)):
            endpos = self.GetTextLength()
            self.GotoPos(endpos)
            event.Skip()
            return

        if canEdit and not self.more and (not shiftDown) and key == wx.WXK_UP:
            # Replace with the previous command from the history buffer.
            self.GoToHistory(True)
        elif canEdit and not self.more and (
                not shiftDown) and key == wx.WXK_DOWN:
            # Replace with the next command from the history buffer.
            self.GoToHistory(False)
        elif canEdit and (not shiftDown) and key == wx.WXK_TAB:
            # show auto-complete list with TAB
            # first try to get the autocompletelist from the package
            cmd = self.getCommand()
            k = self.getAutoCompleteList(cmd)
            cmd = cmd[cmd.rfind('.') + 1:]
            # if failed, search the locals()
            if not k:
                for i in six.moves.range(len(cmd) - 1, -1, -1):
                    if cmd[i].isalnum() or cmd[i] == '_':
                        continue
                    cmd = cmd[i + 1:]
                    break
                k = six.iterkeys(self.interp.locals)
                k = [s for s in k if s.startswith(cmd)]
                k.sort()
            if k:
                self.AutoCompSetAutoHide(self.autoCompleteAutoHide)
                self.AutoCompSetIgnoreCase(self.autoCompleteCaseInsensitive)
                options = ' '.join(k)
                self.AutoCompShow(len(cmd), options)
            return
        else:
            self.searchHistory = True
            super(Shell, self).OnKeyDown(event)

    def OnLeftDClick(self, event):
        line_num = self.GetCurrentLine()
        line = self.GetLine(line_num)
        filepath = re.findall('[Ff]ile [^,]+,', line)
        if filepath:
            path = (filepath[0])[6:-2]
            linenum = re.findall(r'line \d+', line, re.IGNORECASE)
            if linenum:
                linenum = int((linenum[0])[5:])
            else:
                linenum = 1
            dp.send('frame.file_drop',
                    filename=path,
                    activated=True,
                    lineno=linenum)
        event.Skip()

    def GoToHistory(self, up=True):
        """Search up the history buffer for the text in front of the cursor."""
        if not self.CanEdit():
            return
        startpos = self.GetCurrentPos()
        # The text up to the cursor is what we search for.
        numCharsAfterCursor = self.GetTextLength() - startpos
        searchText = self.getCommand(rstrip=False)
        if numCharsAfterCursor > 0:
            searchText = searchText[:-numCharsAfterCursor]
        if not searchText or self.searchHistory == False:
            self.OnHistoryReplace(step=up * 2 - 1)
            self.searchHistory = False
            return
        # Search upwards from the current history position and loop
        # back to the beginning if we don't find anything.
        if up:
            searchOrder = six.moves.range(self.historyIndex + 1,
                                          len(self.history))
        else:
            searchOrder = six.moves.range(self.historyIndex - 1, -1, -1)
        for i in searchOrder:
            command = self.history[i]
            if command[:len(searchText)] == searchText:
                # Replace the current selection with the one we found.
                endpos = self.GetTextLength()
                self.SetSelection(startpos, endpos)
                self.ReplaceSelection(command[len(searchText):])
                endpos = self.GetCurrentPos()
                self.SetSelection(endpos, startpos)
                # We've now warped into middle of the history.
                self.historyIndex = i
                break
            self.historyIndex = i

    def writeOut(self, text):
        """Replacement for stdout."""
        # only output the text when it is not silent
        try:
            if not self.silent:
                # move the cursor to the end to protect the readonly section
                endpos = self.GetTextLength()
                if not self.CanEdit():
                    self.SetCurrentPos(endpos)
                if not self.waiting:
                    # if the shell is in idle status, output the text right before the prompt
                    self.SetCurrentPos(self.promptPosStart)
                    self.write(text)
                    self.promptPosStart += self.GetTextLength() - endpos
                    self.promptPosEnd += self.GetTextLength() - endpos
                else:
                    if self.GetCurrentLine() \
                                    == self.LineFromPosition(self.promptPosEnd):
                        self.write(os.linesep)
                    self.write(text)
                # disable undo
                self.EmptyUndoBuffer()
                # move the caret to the end
                self.GotoPos(self.GetTextLength())
                self.Update()
            else:
                print(text, file=self.stdout)
                #for line in traceback.format_stack():
                #    print(line.strip(), file=self.stdout)
        except:
            print(text, file=self.stdout)
            traceback.print_exc(file=self.stdout)

    def writeErr(self, text):
        """Replacement for stderror"""
        self.writeOut(text)

    def addHistory(self, command):
        # override the parent function to add time-stamp.
        stamp = time.strftime('#bsm#%Y/%m/%d')
        if stamp not in self.history:
            self.history.insert(0, stamp)
        super(Shell, self).addHistory(command)

    def runCommand(self,
                   command,
                   prompt=True,
                   verbose=True,
                   debug=False,
                   history=True):
        """run the command in the shell"""
        if not self.enable_debugger:
            self.enable_debugger = debug
        self.autoIndent = False
        startpos = self.promptPosEnd
        endpos = self.GetTextLength()
        # Go to the very bottom of the text.
        self.SetCurrentPos(endpos)

        # save the currently typed command
        command_typed = ""
        if not self.running and not self.more:
            command_typed = self.GetTextRange(startpos, endpos)
            self.clearCommand()

        command = command.rstrip()
        if verbose:
            self.write(command)
        self.push(command, not prompt, history)

        # retrieve the typed command
        if not self.more and command_typed:
            self.write(command_typed)
        self.autoIndent = True

    def push(self, command, silent=False, history=True):
        """Send command to the interpreter for execution."""
        self.running = True
        if not silent:
            self.write(os.linesep)
        # push to the debugger
        if self.waiting and self.IsDebuggerOn():
            self.debugger.push_line(command)
            return
        # DNM
        cmd_raw = command
        if pyshell.USE_MAGIC:
            command = magic(command)
        if len(command) > 1 and command[-1] == ';':
            self.silent = True

        self.waiting = True
        self.lastUpdate = None
        try:
            if self.enable_debugger:
                self.debugger.reset()
                sys.settrace(self.debugger)
            self.more = self.interp.push(command)
        except:
            traceback.print_exc(file=sys.stdout)
        finally:
            # make sure debugger.ended is always sent; more does not hurt
            if self.enable_debugger:
                dp.send('debugger.ended')
                self.debugger.reset()
                self.enable_debugger = False

        sys.settrace(None)
        self.lastUpdate = None
        self.waiting = False
        self.silent = False
        if not self.more and history:
            self.addHistory(cmd_raw)
        if not silent:
            self.prompt()
        self.running = False

    def lstripPrompt(self, text):
        """Return text without a leading prompt."""
        ps = [str(sys.ps1), str(sys.ps2), str(sys.ps3), str("K>> ")]
        # Strip the prompt off the front of text.
        for p in ps:
            if text[:len(p)] == p:
                text = text[len(p):]
                break
        return text

    def prompt(self):
        """Display proper prompt for the context: ps1, ps2 or ps3.

        If this is a continuation line, autoindent as necessary."""
        isreading = self.reader.isreading
        skip = False
        if isreading:
            prompt = str(sys.ps3)
        elif self.more:
            prompt = str(sys.ps2)
        elif self.waiting and self.IsDebuggerOn():
            prompt = 'K>> '
        else:
            prompt = str(sys.ps1)
        pos = self.GetCurLine()[1]
        if pos > 0:
            if isreading:
                skip = True
            else:
                self.write(os.linesep)
        if not self.more:
            self.promptPosStart = self.GetCurrentPos()
        if not skip:
            # no need to go to the end position
            #endpos = self.GetTextLength()
            #self.GotoPos(endpos)
            self.write(prompt)
        if not self.more:
            self.promptPosEnd = self.GetCurrentPos()
            # Keep the undo feature from undoing previous responses.
            self.EmptyUndoBuffer()
        if self.more:
            line_num = self.GetCurrentLine()
            currentLine = self.GetLine(line_num)
            previousLine = self.GetLine(line_num - 1)[len(prompt):]
            pstrip = previousLine.strip()
            lstrip = previousLine.lstrip()
            # Get the first alnum word:
            first_word = []
            for i in pstrip:
                if i.isalnum():
                    first_word.append(i)
                else:
                    break
            first_word = ''.join(first_word)
            if pstrip == '':
                # because it is all whitespace!
                indent = previousLine.strip('\n').strip('\r')
            else:
                indent = previousLine[:len(previousLine) - len(lstrip)]
                keys = [
                    'if', 'else', 'elif', 'for', 'while', 'def', 'class',
                    'try', 'except', 'finally'
                ]
                if pstrip[-1] == ':' and first_word in keys:
                    indent += ' ' * 4
            if self.autoIndent:
                self.write(indent)
        self.SetSavePoint()
        self.EnsureCaretVisible()
        self.ScrollToColumn(0)

    def autoCallTipShow(self, command, insertcalltip=True, forceCallTip=False):
        """Display argument spec and docstring in a popup window."""
        if self.CallTipActive():
            self.CallTipCancel()
        (name, argspec, tip) = self.interp.getCallTip(command)
        if tip:
            dp.send('Shell.calltip', sender=self, calltip=tip)
        if not self.autoCallTip and not forceCallTip:
            return
        startpos = self.GetCurrentPos()
        if argspec and insertcalltip and self.callTipInsert:
            self.write(argspec + ')')
            endpos = self.GetCurrentPos()
            self.SetSelection(startpos, endpos)
        if argspec:
            tippos = startpos
            fallback = startpos - self.GetColumn(startpos)
            # In case there isn't enough room, only go back to the
            # fallback.
            tippos = max(tippos, fallback)
            self.CallTipShow(tippos, argspec)

    def debugPrompt(self, ismore=False, iserr=False):
        """show the debug prompt"""
        startpos = self.promptPosEnd
        endpos = self.GetTextLength()
        self.GotoPos(endpos)
        self.more = ismore
        # skip the uncessary prompt, it is not ideal and will eat the typed
        # commands
        if self.GetCurLine()[0] == "K>> ":
            return
        autoIndent = self.autoIndent
        if self.more:
            self.autoIndent = True
        self.prompt()
        self.autoIndent = autoIndent
        return

    def processLine(self):
        """Process the line of text at which the user hit Enter."""

        # The user hit ENTER and we need to decide what to do. They
        # could be sitting on any line in the shell.

        thepos = self.GetCurrentPos()
        startpos = self.promptPosEnd
        endpos = self.GetTextLength()
        ps2 = str(sys.ps2)
        # If they hit RETURN inside the current command, execute the
        # command.
        if self.CanEdit():
            execute = self.GetCurrentLine() == self.LineFromPosition(endpos)
            if self.more and not execute:
                # if the interpreter needs more code and the current position
                # is not on last line, insert a line after current position to
                # allow the user inputs more code
                self.prompt()
                return
            self.SetCurrentPos(endpos)
            self.interp.more = False
            command = self.GetTextRange(startpos, endpos)
            lines = command.split(os.linesep + ps2)
            lines = [line.rstrip() for line in lines]
            command = '\n'.join(lines)
            if self.reader.isreading:
                if not command:
                    # Match the behavior of the standard Python shell
                    # when the user hits return without entering a
                    # value.
                    command = '\n'
                self.reader.input = command
                self.write(os.linesep)
            else:
                self.push(command)
                wx.CallAfter(self.EnsureCaretVisible)
        # Or replace the current command with the other command.
        else:
            # If the line contains a command (even an invalid one).
            if self.getCommand(rstrip=False):
                command = self.getMultilineCommand()
                self.clearCommand()
                self.write(command)
            # Otherwise, put the cursor back where we started.
            else:
                self.SetCurrentPos(thepos)
                self.SetAnchor(thepos)

    def setStyles(self, faces):
        super().setStyles(faces)
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER,
                          'fore:#3C3C43,back:#F2F2F7')

    @classmethod
    def Initialize(cls, frame, **kwargs):
        cls.frame = frame
        dp.connect(receiver=cls.Uninitialize, signal='frame.exit')
        ns = {}
        ns['wx'] = wx
        ns['app'] = wx.GetApp()
        ns['frame'] = cls.frame
        intro = 'Welcome To bsmedit ' + __version__
        cls.panelShell = Shell(cls.frame, introText=intro, locals=ns)
        active = kwargs.get('active', True)
        direction = kwargs.get('direction', 'top')
        dp.send(signal="frame.add_panel",
                panel=cls.panelShell,
                active=active,
                title="Shell",
                showhidemenu="View:Panels:Console",
                direction=direction)

    @classmethod
    def Uninitialize(cls):
        if cls.panelShell:
            dp.send('frame.delete_panel', panel=cls.panelShell)


def bsm_initialize(frame, **kwargs):
    Shell.Initialize(frame, **kwargs)



from .sim import sim, SimPropGrid


def gcp():
    """
    get the current propgrid manager
    """
    mgr = SimPropGrid.GCM.get_active()
    if not mgr:
        mgr = sim.propgrid()
    return mgr


# add some shortcuts
progrid = sim.propgrid
simulation = sim.simulation
plot_trace = sim.plot_trace



import wx
import wx.lib.agw.aui as aui
import wx.py as py
from auibarpopup import *

ID_TOOL_START = wx.ID_HIGHEST + 1

class MyPanel(wx.Panel):
    def __init__(self, parent, *args, **kwargs):
        wx.Panel.__init__(self, parent, *args, **kwargs)
        self.toolbarart = AuiToolBarPopupArt(self)
        sizer = wx.BoxSizer(wx.VERTICAL)
        tb = aui.AuiToolBar(self, -1, wx.DefaultPosition, wx.DefaultSize,
                            agwStyle=aui.AUI_TB_OVERFLOW|aui.AUI_TB_TEXT|
                            aui.AUI_TB_HORZ_TEXT|aui.AUI_TB_PLAIN_BACKGROUND)
        tb.SetToolBitmapSize(wx.Size(16, 16))
        tb_bmp1 = wx.ArtProvider.GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, wx.Size(16, 16))
        tb.AddSimpleTool(ID_TOOL_START+200, "Item 1", tb_bmp1)
        tb.AddCheckTool(ID_TOOL_START+201, "Toggle", tb_bmp1, wx.NullBitmap)
        tb.AddRadioTool(ID_TOOL_START+202, "Radio 0", tb_bmp1, wx.NullBitmap)
        tb.AddRadioTool(ID_TOOL_START+203, "Radio 1", tb_bmp1, wx.NullBitmap)
        tb.AddSeparator()
        tb.AddSimpleTool(ID_TOOL_START+204, "Item 5", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+205, "Item 6", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+206, "Item 7", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+207, "Item 8", tb_bmp1)
        tb.SetArtProvider(self.toolbarart)
        tb.Realize()
        sizer.Add(tb, 0, wx.EXPAND)
        ns = {}
        ns['wx'] = wx
        ns['app'] = wx.GetApp()
        ns['frame'] = parent
        shell = py.shell.Shell(self, -1, locals=ns)
        sizer.Add(shell, 1, wx.EXPAND)
        self.SetSizer(sizer)

        self.Bind(wx.EVT_TOOL, self.OnTool)

    def OnTool(self, event):
        print('ID %d called in panel'%(event.GetId()))

class MainFrame(wx.Frame):
    def __init__(self):
        wx.Frame.__init__(self, None, -1, 'AuiToolBarPopup Demo', size=(300, 432))
        self._mgr = aui.AuiManager()

        # tell AuiManager to manage this frame
        self._mgr.SetManagedWindow(self)
        self.toolbarart = AuiToolBarPopupArt(self)

        self._mgr.AddPane(MyPanel(self), aui.AuiPaneInfo().Name("test1").
                          Caption("Pane Caption").CenterPane())

        tb = aui.AuiToolBar(self, -1, wx.DefaultPosition, wx.DefaultSize,
                            agwStyle=aui.AUI_TB_OVERFLOW|aui.AUI_TB_TEXT|
                            aui.AUI_TB_HORZ_TEXT|aui.AUI_TB_PLAIN_BACKGROUND)
        tb.SetToolBitmapSize(wx.Size(16, 16))
        tb_bmp1 = wx.ArtProvider.GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, wx.Size(16, 16))
        tb.AddSimpleTool(ID_TOOL_START+0, "Item 1", tb_bmp1)
        tb.AddCheckTool(ID_TOOL_START+1, "Toggle", tb_bmp1, wx.NullBitmap)
        tb.AddSeparator()
        tb.AddRadioTool(ID_TOOL_START+2, "Radio 0", tb_bmp1, wx.NullBitmap)
        tb.AddRadioTool(ID_TOOL_START+3, "Radio 1", tb_bmp1, wx.NullBitmap)
        tb.AddSeparator()
        tb.AddSimpleTool(ID_TOOL_START+4, "Item 5", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+5, "Item 6", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+6, "Item 7", tb_bmp1)
        tb.AddSimpleTool(ID_TOOL_START+7, "Item 8", tb_bmp1)

        choice = wx.Choice(tb, -1, choices=["One choice", "Another choice"])
        tb.AddControl(choice)
        cmb = wx.ComboBox(tb, -1, choices=["Option 1", "Option 2", "Option 3"])
        tb.AddControl(cmb)
        spin = wx.SpinCtrl(tb, -1)
        tb.AddControl(spin)

        tb.SetToolDropDown(ID_TOOL_START+6, True)
        tb.Realize()
        tb.SetArtProvider(self.toolbarart)
        self._mgr.AddPane(tb, aui.AuiPaneInfo().Name("tb").Caption("tb").
                          ToolbarPane().Floatable(False).Top())

        tb2 = aui.AuiToolBar(self, -1, wx.DefaultPosition, wx.DefaultSize,
                             agwStyle=aui.AUI_TB_OVERFLOW|aui.AUI_TB_TEXT|
                             aui.AUI_TB_HORZ_TEXT|aui.AUI_TB_PLAIN_BACKGROUND)
        tb2.SetToolBitmapSize(wx.Size(16, 16))
        tb2.AddSimpleTool(ID_TOOL_START+100, "Item 1", tb_bmp1)
        tb2.AddSimpleTool(ID_TOOL_START+101, "Item 2", tb_bmp1)
        tb2.AddSimpleTool(ID_TOOL_START+102, "Item 3", tb_bmp1)
        tb2.AddSimpleTool(ID_TOOL_START+103, "Item 4", tb_bmp1)
        tb2.AddSeparator()
        statictext = wx.StaticText(tb2, -1, "StaticText")
        tb2.AddControl(statictext)
        slider = wx.Slider(tb2, 100, 25, 1, 100, size=(250, -1), style=wx.SL_HORIZONTAL)
        tb2.AddControl(slider)
        search = wx.SearchCtrl(tb2, size=(200, -1), style=wx.TE_PROCESS_ENTER)
        tb2.AddControl(search)
        tb2.Realize()
        tb2.SetArtProvider(self.toolbarart)
        self._mgr.AddPane(tb2, aui.AuiPaneInfo().Name("tb2").Caption("tb2").
                          ToolbarPane().Floatable(False).Top().Row(1))
        self._mgr.Update()
        self.Bind(wx.EVT_TOOL, self.OnTool)
        self.Bind(aui.EVT_AUITOOLBAR_TOOL_DROPDOWN, self.OnDropDownToolbarItem,
                  id=ID_TOOL_START+6)

    def OnTool(self, event):
        print('ID %d called'%(event.GetId()))

    def OnDropDownToolbarItem(self, event):
        if event.IsDropDownClicked():
            tb = event.GetEventObject()
            print("drop down, ID:%d"%(event.GetId()))
            tb.SetToolSticky(event.GetId(), True)

            # create the popup menu
            menuPopup = wx.Menu()
            bmp = wx.ArtProvider.GetBitmap(wx.ART_QUESTION, wx.ART_OTHER, wx.Size(16, 16))

            m1 = wx.MenuItem(menuPopup, 10001, "Drop Down Item 1")
            m1.SetBitmap(bmp)
            menuPopup.Append(m1)

            m2 = wx.MenuItem(menuPopup, 10002, "Drop Down Item 2")
            m2.SetBitmap(bmp)
            menuPopup.Append(m2)

            m3 = wx.MenuItem(menuPopup, 10003, "Drop Down Item 3")
            m3.SetBitmap(bmp)
            menuPopup.Append(m3)

            m4 = wx.MenuItem(menuPopup, 10004, "Drop Down Item 4")
            m4.SetBitmap(bmp)
            menuPopup.Append(m4)

            # line up our menu with the button
            rect = tb.GetToolRect(event.GetId())
            pt = tb.ClientToScreen(rect.GetBottomLeft())
            pt = self.ScreenToClient(pt)

            self.PopupMenu(menuPopup, pt)

            # make sure the button is "un-stuck"
            tb.SetToolSticky(event.GetId(), False)

class RunApp(wx.App):
    def __init__(self):
        wx.App.__init__(self, redirect=False)

    def OnInit(self):
        frame = MainFrame()
        frame.Show(True)
        self.SetTopWindow(frame)
        self.frame = frame
        return True

def main():
    app = RunApp()
    app.MainLoop()

if __name__ == '__main__':
    main()



# https://github.com/ActiveState/code/blob/master/recipes/Python/305271_Enumerated_values_name_or/LICENSE.md
# Originally published: 2004-09-13 16:10:21
# Last updated: 2004-09-13 16:10:21
# Author: Samuel Reynolds
#
# This enumerated-values class is intended for use in module-level constants,
# such as an employee type or a status value. It allows simple attribute
# notation (e.g., employee.Type.Serf or employee.Type.CEO), but also supports
# simple bidirectional mapping using dictionary notation (e.g.,
# employee.Type['CEO']-->1 or employee.Type[1]-->'CEO').

# Copyright 2004 Samuel Reynolds
#
# The software is licensed according to the terms of the PSF (Python Software
# Foundation) license found here: http://www.python.org/psf/license/

import six


class EnumType(object):
    """
    Enumerated-values class.

    Allows reference to enumerated values by name
    or by index (i.e., bidirectional mapping).
    """
    def __init__(self, *args, **kwargs):
        # Remember names list for reference by index
        self._items = {}
        # Attributes for direct reference
        for _i, _s in enumerate(args):
            self._items[_i] = _s
            setattr(self, _s, _i)
        for _s, _i in six.iteritems(kwargs):
            self._items[_i] = _s
            setattr(self, _s, _i)

    def __contains__(self, item):
        try:
            trans = self[item]
            return True
        except:
            return False

    def __iter__(self):
        return self.items()

    def __getitem__(self, key):
        if key in self._items:
            return self._items[key]
        else:
            return self._nameToEnum(key)

    def __len__(self):
        return len(self._items)

    def items(self):
        return list(six.iteritems(self._items))

    def names(self):
        return list(self._items.keys())

    def _nameToEnum(self, name):
        try:
            return getattr(self, name)
        except ValueError as exc:
            args = list(exc.args)
            args.append("Unknown enum value name '%s'" % name)
            args = tuple(args)
            exc.args = args
            raise



"""define some utility functions"""
import six
import wx
import wx.lib.agw.aui as aui
import wx.svg

def MakeBitmap(red, green, blue, alpha=128):
    # Create the bitmap that we will stuff pixel values into using
    # the raw bitmap access classes.
    bmp = wx.Bitmap(16, 16, 32)

    # Create an object that facilitates access to the bitmap's
    # pixel buffer
    pixelData = wx.AlphaPixelData(bmp)
    if not pixelData:
        raise RuntimeError("Failed to gain raw access to bitmap data.")

    # We have two ways to access each pixel, first we'll use an
    # iterator to set every pixel to the colour and alpha values
    # passed in.
    for pixel in pixelData:
        pixel.Set(red, green, blue, alpha)

    # Next we'll use the pixel accessor to set the border pixels
    # to be fully opaque
    pixels = pixelData.GetPixels()
    for x in six.moves.range(16):
        pixels.MoveTo(pixelData, x, 0)
        pixels.Set(red, green, blue, wx.ALPHA_OPAQUE)
        pixels.MoveTo(pixelData, x, 16 - 1)
        pixels.Set(red, green, blue, wx.ALPHA_OPAQUE)
    for y in six.moves.range(16):
        pixels.MoveTo(pixelData, 0, y)
        pixels.Set(red, green, blue, wx.ALPHA_OPAQUE)
        pixels.MoveTo(pixelData, 16 - 1, y)
        pixels.Set(red, green, blue, wx.ALPHA_OPAQUE)

    return bmp


def PopupMenu(wnd, menu):
    # popup a menu, and return the selected command
    if not wnd or not menu:
        return wx.ID_NONE

    cc = aui.ToolbarCommandCapture()
    wnd.PushEventHandler(cc)

    wnd.PopupMenu(menu)

    command = cc.GetCommandId()
    wnd.PopEventHandler(True)
    return command


class FastLoadTreeCtrl(wx.TreeCtrl):
    """
    When a treectrl tries to load a large amount of items, it will be slow.
    This class will not load the children item until the parent is expanded (
    e.g., by a click).
    """
    def __init__(self,
                 parent,
                 getchildren=None,
                 style=wx.TR_DEFAULT_STYLE,
                 sort=True):
        wx.TreeCtrl.__init__(self, parent, style=style)
        self._get_children = getchildren
        assert self._get_children
        self._sort_children = sort
        self.Bind(wx.EVT_TREE_ITEM_EXPANDING, self.OnTreeItemExpanding)

    def OnTreeItemExpanding(self, event):
        """expand the item with children"""
        item = event.GetItem()
        if not item.IsOk():
            return
        self.FillChildren(item)

    def FillChildren(self, item):
        """fill the node with children"""
        if not ((self.GetWindowStyle() & wx.TR_HIDE_ROOT)
                and item == self.GetRootItem()):
            child, _ = self.GetFirstChild(item)
            if not child.IsOk():
                return False
            if self.GetItemText(child) != "...":
                return False
        # delete the '...'
        self.DeleteChildren(item)
        children = self._get_children(item)
        for obj in children:
            # fill all the children
            child = self.AppendItem(item, obj['label'], obj['img'],
                                    obj['imgsel'], obj['data'])
            # add the place holder for children
            if obj['is_folder']:
                self.AppendItem(child, '...', -1, -1, None)
            clr = obj.get('color', None)
            if clr:
                self.SetItemTextColour(child, wx.Colour(100, 174, 100))
        if self._sort_children:
            self.SortChildren(item)
        return True

def svg_to_bitmap(svg, width=16, height=16):
    bmp = wx.svg.SVGimage.CreateFromBytes(str.encode(svg))
    return bmp.ConvertToScaledBitmap((width, height))
